{ $Header: /cvsroot/InfExp/InfExp/src/InfStruc.pas,v 1.13 2000/11/01 19:00:14 yole Exp $
  Infinity Engine data structures
  Copyright (C) 2000-02 Dmitry Jemerov <yole@yole.ru>
  See the file COPYING for license information
}

unit InfStruc;

interface

{$ALIGN OFF}

type
  TFileType = (ftNone, ftBMP, ftMVE, ftWAV, ftWFX, ftPLT, ftBAM, ftWED,
       ftCHU, ftMAP, ftMOS, ftITM, ftSPL, ftSCRIPT, ftIDS, ftCRE, ftAREA,
       ftDLG, ft2DA, ftGAME, ftSTOR, ftWMAP, ftEFF, ftVVC, ftPRO, ftINI,
       ftSRC);
  TFileTypes = set of TFileType;

  TResRef = array [0..7] of char;

  TGameType = (gtBaldur, gtTorment, gtIcewind, gtIcewind2);

// -- CHITIN.KEY and BIFFs ---------------------------------------------------

  TChitinKeyHdr = record
    Signature: array [0..7] of char;
    BIFF_Count: LongInt;
    File_Count: LongInt;
    BIFF_Offset: LongInt;
    File_Offset: LongInt;
  end;

  TChitinBIFFRec = record
    BIFF_Length: LongInt;
    Name_Offset: LongInt;
    Name_Length: word;
    CD_Mask: word;
  end;

  TChitinFileRec = record
    Name: array [0..7] of char;
    FileType: word;
    Index: word;
    BIFF: word;
  end;

  TBIFFHeaderRec = record
    Signature: array [0..7] of char;
    ElemCount, MapCount: integer;
    DirOffset: integer;
  end;

  TBIFFEntryRec = record
    ElemIndex, BIFFIndex: word;
    Offset, Size, FileType: integer;
  end;

  TBIFFMapEntryRec = record
    MapIndex: LongInt;
    Offset: integer;
    BlockCount, BlockSize: integer;
    FileType: integer;
  end;

// -- TLK --------------------------------------------------------------------

  TTLKHeaderRec = record
    Signature: array [0..7] of char;
    Unknown: word;
    RecordCount, TextOffset: LongInt;
  end;

  TTLKEntryRec = record
    Flag: word;
    Sound: array [0..15] of char;
    Offset, Length: LongInt;
  end;

// -- DLG --------------------------------------------------------------------

  TDlgHeaderRec = record
    Signature: array [0..7] of char;
    PhraseCount, PhraseOffset: LongInt;
    ResponceCount, ResponceOffset: LongInt;
    CondOffset, CondCount: LongInt;
    RespCondOffset, RespCondCount: LongInt;
    ActionOffset, ActionCount: LongInt;
  end;

  TDlgPhraseRec = record
    TlkIndex: LongInt;
    FirstResponce, ResponceCount: LongInt;
    Cond: LongInt;
  end;

  TDlgResponceRec = record
    Flags: LongInt;
    TlkIndex: LongInt;
    JournalIndex: LongInt;
    Cond, Action: LongInt;
    Dlg: TResRef;
    DlgPhrase: LongInt;
  end;

// -- MOS --------------------------------------------------------------------

  TMOSHeaderRec = record
    Signature: array [0..7] of char;
    Width, Height: word;
    Cols, Rows: word;
    BlkSize, BlkBPP: LongInt;
  end;

  TPaletteEntry = record
    R, G, B, A: byte;
  end;

  TImagePalette = array [0..255] of TPaletteEntry;

// -- BAM --------------------------------------------------------------------

  TBAMHeaderRec = record
    Signature: array [0..7] of char;
    FrameCnt: word;
    AnimCnt: byte;
    TransparentColor: byte;
    FrameHeaderPtr: LongInt;
    PalettePtr: LongInt;
    AnimArrayPtr: LongInt;
  end;

  TBAMFrameHeaderRec = record
    Width, Height: word;
    XPos, YPos: SmallInt;
    GraphicPtr: LongWord;
  end;

  TBAMAnimHeaderRec = record
    Frames, Start: word;
  end;

// -- WED --------------------------------------------------------------------

  TWEDHeaderRec = record
    Signature: array [0..7] of char;
    EnvCount: LongInt;
    DoorCount: LongInt;
    EnvOffset: LongInt;
    OffsetsOffset: LongInt;
    DoorOffset: LongInt;
    DoorTileCellOffset: LongInt;
  end;

  TWEDEnvRec = record
    XBlocks, YBlocks: word;
    ResourceID: TResRef;
    Unknown: LongInt;
    TileMapOffset: LongInt;
    TileIndexLookupOffset: LongInt;
  end;

  TWEDOffsetsRec = record
    WallPolyCount: LongInt;
    WallPolyOffset: LongInt;
    PointOffset: LongInt;
    WallGroupOffset: LongInt;
    IndexOffset: LongInt;
  end;

  TWEDDoorRec = record
    Id: array [0..7] of char;
    Unknown: array [1..4] of word;
    PolyCount: word;
    PolyOffset1, PolyOffset2: LongInt;
  end;

  TWEDPolyRec = record
    FirstPoint, PointCount: LongInt;
    Passable: byte;
    Unknown: byte;
    XMin, XMax, YMin, YMax: word;
  end;

  TWEDTileRec = record
    PrimaryTileIndex: word;
    PrimaryTileCount: word;
    SecondaryTileIndex: word;
    OverlayMask: byte;
    Unknown: array [1..3] of byte;
  end;

// -- ARE --------------------------------------------------------------------

  TAreaHeaderRec = record
    Signature: array [0..7] of char;
    AreaID: TResRef;
    Unknown1: array [1..8] of byte;
    ExitN: TResRef;
    Unknown2: LongInt;
    ExitW: TResRef;
    Unknown3: LongInt;
    ExitS: TResRef;
    Unknown4: LongInt;
    ExitE: TResRef;
    Unknown5: LongInt;
    Flags: word;
    RainProbability, SnowProbability: word;
    FogProbability, LightningProbability: word;
    Unknown6: word;
  end;

  TAreaOffsetsRec = record
    NPCOffset: LongInt;
    NPCCount: word;
    InfoCount: word;
    InfoOffset: LongInt;
    SpawnPointOffset, SpawnPointCount: LongInt;
    EntryOffset: LongInt;
    EntryCount: LongInt;
    ContainerOffset: LongInt;
    ContainerCount: word;
    ItemCount: word;
    ItemOffset: LongInt;
    PointOffset: LongInt;
    PointCount: word;
    SoundCount: word;
    SoundOffset: LongInt;
    VarOffset: LongInt;
    VarCount: LongInt;
    Unknown7: LongInt;
    AIScript: array [0..7] of char;
    Unknown8: array [1..8] of byte;
    DoorCount, DoorOffset: LongInt;
    AnimCount, AnimOffset: LongInt;
    TiledObjectCount, TiledObjectOffset: LongInt;
  end;

  TAreaNPCRec = record
    Name: array [0..31] of char;
    CurX, CurY: word;
    DestX, DestY: word;
    Visible: LongInt;
    Unknown1: array [1..28] of byte;
    Dialog: TResRef;
    ScriptOverride: TResRef;
    ScriptClass: TResRef;
    ScriptRace: TResRef;  
    ScriptGeneral: TResRef;
    ScriptDefault: TResRef;
    ScriptSpecifics: TResRef;
    CREfile: TResRef;
    CREoffset, CREsize: LongInt;
    Unknown: array [1..128] of byte;
  end;

  TAreaInfoRec = record
    Id: array [0..31] of char;
    TriggerType: word;
    X1, Y1, X2, Y2: word;
    PointCount: word;
    FirstPoint: LongInt;
    Unknown1: LongInt;
    CursorIndex: LongInt;
    DestArea: TResRef;
    DestEntrance: array [0..31] of char;
    Flags: LongInt;
    DescTLKIndex: LongInt;
    TrapDetectionDifficulty: word;
    TrapRemovalDifficulty: word;
    Trapped, TrapDetected: word;
    TrapLaunchX, TrapLaunchY: word;
    KeyType: TResRef;
    Script: TResRef;
    Unknown2: array [1..28] of word;
    Dialog: TResRef;
  end;

  TAreaTriggerType = (attProximity, attInfo, attTravel);

const
  atfResetTrap = 2;
  atfPartyRequired = 4;

type
  TAreaEntryRec = record
    Name: array [0..31] of char;
    X, Y: word;
    Unknown: array [1..68] of byte;
  end;

  TAreaContainerRec = record
    Name: array [0..31] of char;
    X, Y: word;
    ContainerType: word;
    LockDifficulty: word;
    Locked: word;
    Unknown: word;
    TrapDetectionDifficulty: word;
    TrapRemovalDifficulty: word;
    Trapped, TrapDetected: word;
    TrapLaunchX, TrapLaunchY: word;
    X1, Y1, X2, Y2: word;
    StartItem, ItemCount: LongInt;
    TrapScript: TResRef;
    FirstPoint, PointCount: LongInt;
    Unknown2: array [1..32] of char;
    KeyType: TResRef;
    Unknown3: array [1..64] of byte;
  end;

  TAreaDoorRec = record
    Name: array [0..31] of char;
    Id: array [0..7] of char;
    Flags: LongInt;
    ClosedFirstPoint: LongInt;
    ClosedPointCount, OpenPointCount: word;
    OpenFirstPoint: LongInt;
    ClosedX1, ClosedY1, ClosedX2, ClosedY2: word;
    OpenX1, OpenY1, OpenX2, OpenY2: word;
    Unknown2: array [1..36] of byte;
    TrapDetectionDifficulty: word;
    TrapRemovalDifficulty: word;
    Trapped, TrapDetected: word;
    TrapLaunchX, TrapLaunchY: word;
    KeyType: TResRef;
    TrapScript: TResRef;
    Unknown4: LongInt;
    LockDifficulty: LongInt;
    X1, Y1, X2, Y2: word;   // unknown
    Unknown5: array [1..48] of byte;
  end;

  TAreaTiledObjectRec = record
    Name: array [0..31] of char;
    Id: array [0..7] of char;
    Unknown: LongInt;
    PrimarySearchSquaresStart, PrimarySearchSquaresCount: integer;
    SecondarySearchSquaresStart, SecondarySearchSquaresCount: integer;
    Unknown2: array [1..48] of byte;
  end;

  TAreaAnimRec = record
    Name: array [0..31] of char;
    X, Y: word;
    Unknown: LongInt;
    BAMId: TResRef;
    BAMCycle: word;
    Unknown2: array [1..26] of byte;
  end;

  TAreaSpawnRec = record
    Name: array [0..31] of char;
    X, Y: word;
    Creatures: array [0..9] of TResRef;
    CreatureCount: word;
    Unknown: array [1..2] of byte;
    field_78: word;
    Flags: word;
    Unknown2: array [1..8] of byte;
    field_84: word;
    field_86: word;
    field_88: LongInt;
    Probability1: word;
    Probability2: word;
    Unknown3: array [1..56] of byte;
  end;

// -- ITM --------------------------------------------------------------------

  TItem10HeaderRec = record
    Signature: array [0..7] of char;
    NameUnident, NameIdent: LongInt;
    Destroyed: TResRef;
    Attr: LongInt;
    ItemType: word;
    ItemUsability: LongInt;
    WeaponID: array [0..1] of char;
    MinLevel: word;
    MinStr: byte;
    Unknown1: byte;
    MinStrBonus: byte;
    Unknown2: byte;
    MinInt: byte;
    Unknown3: byte;
    MinDex: byte;
    Unknown4: byte;
    MinWis: byte;
    Unknown5: byte;
    MinCon: byte;
    Unknown6: byte;
    MinCha: byte;
    Unknown7: byte;
    Price: LongInt;
    MaxInStack: word;
    BAMIdInv: TResRef;
    LoreToIdentify: word;
    BAMIdGround: TResRef;
    Weight: LongInt;
    DescUnident, DescIdent: LongInt;
    BAMIdCarried: TResRef;
    Enchanted: LongInt;
    AbilityPtr: LongInt;
    AbilityCount: word;
    EffectTablePtr: LongInt;
    FirstGlobalEffect: word;
    GlobalEffectCount: word;
  end;

  // in V1.1 items, goes after 1.0 header
  TItem11ExHeaderRec = record
    Dialog: TResRef;
    UnknownTLK: LongInt;
    Unknown4: array [1..14] of SmallInt;
  end;

  TItem11HeaderRec = record
    Hdr10: TItem10HeaderRec;
    HdrEx: TItem11ExHeaderRec;
  end;

  TAbilityRec = record
    AbilityType: byte;
    UseOnlyIdentified: byte;
    Unknown1: word;
    BAMIcon: TResRef;
    TargetType: word;
    Range: word;
    LauncherType: word;
    Speed: word;
    Bonus: SmallInt;
    DiceValue: word;
    DiceCount: word;
    DmgBonus: SmallInt;
    DamageType: word;
    EffectCount: word;
    FirstEffect: word;
    ChargeCount: word;
    Unknown6: array [1..2] of word;
  end;

  TSpellAbilityRec = record
    AbilityType: byte;
    UseOnlyIdentified: byte;
    Unknown1: word;
    BAMIcon: TResRef;
    TargetType: word;
    Range: word;
    MinLevel: word;
    Speed: word;
    Bonus: SmallInt;
    DiceValue: word;
    DiceCount: word;
    DmgBonus: SmallInt;
    DamageType: word;
    EffectCount: word;
    FirstEffect: word;
    Unknown6: array [1..2] of word;
    ProjectileType: word;
  end;

  TItemAbilityRec = record
    Base: TAbilityRec;
    Unknown: word;
    ProjectileType: word;
    Unknown2: array [1..3] of word;
    IsArrow, IsBolt, IsMissile: word;
  end;

  TEffectRec = record
    EffectType: word;
    TargetType: byte;
    Unknown: byte;
    NP1, NP2: LongInt;              // number parameters
    Flags: word;
    Time: LongInt;
    Prob1: byte;                     // probability
    Prob2: byte;
    Resource: TResRef;
    DiceCount: LongInt;
    DiceValue: LongInt;
    SaveType: LongInt;
    SaveBonus: LongInt;
    Unknown2: LongInt;
  end;
  PEffectRec = ^TEffectRec;

  TEffect20Rec = record
    Signature: array [0..7] of char;
    EffectType: LongInt;
    TargetType: LongInt;
    Unknown: LongInt;
    NP1, NP2: LongInt;
    Flags: LongInt;
    Time: LongInt;
    Prob1: word;
    Prob2: word;
    Resource: TResRef;
    DiceValue: LongInt;
    DiceCount: LongInt;
    SaveType: LongInt;
    SaveBonus: LongInt;
    Unknown2: array [1..3] of LongInt;
    DiceValue2: LongInt;
    DiceCount2: LongInt;
    Flags2: LongInt;
    Unknown3: array [1..4] of LongInt;
    UnknownResRef: array [1..2] of TResRef;
    Unknown4: array [1..5] of LongInt;
    UnknownResRef2: TResRef;
    Unknown5: array [1..116] of byte;
  end;

// -- CRE --------------------------------------------------------------------

  TSpellType = (stInvalid, stWizard, stPriest, stInnate);

  TCRECommon0Rec = record
    Signature: array [0..7] of char;
    NameTLK, TooltipTLK: integer;
    Flags: integer;
    KillXP, CurXP: integer;
    Gold: integer;
    Status: integer;
    MaxHP, CurHP: word;
    AnimID: word;
    Unknown2: word;
    ColorMetal: byte;
    ColorMinor, ColorMajor, ColorSkin: byte;
    ColorLeather, ColorArmor: byte;
    ColorHair: byte;
    EffectFormat: byte;
    ImageS, ImageL: TResRef;
    Reputation: byte;
    Unknown6: byte;
    BaseAC: SmallInt;
    AC: SmallInt;
    ACCrushing: SmallInt;
    ACMissile: SmallInt;
    ACPiercing: SmallInt;
    ACSlashing: SmallInt;
    THAC0: ShortInt;
    NumberOfAttacks: byte;
    ST_PPDM, ST_RSW, ST_PP, ST_BW, ST_Spells: ShortInt;
    ResistFire, ResistCold, ResistElec, ResistAcid: ShortInt;
    ResistMagic, ResistMagFire, ResistMagCold: ShortInt;
    ResistSlashing, ResistPiercing, ResistCrushing, ResistMissile: ShortInt;
    Unknown7: array [1..2] of byte;
    Lore: byte;
    OpenLocks, Stealth, DetectTraps, PickPockets: byte;
    Fatigue, Intoxication: byte;
    Luck: ShortInt;
  end;

  TCRECommon1Rec = record
    WeaponProf: array [1..15] of byte;
    Unknown1: array [1..6] of byte;
    Tracking: byte;
    Unknown8a: array [1..32] of byte;
    TLKIndex: array [1..100] of LongInt;
    Level: array [1..3] of ShortInt;
    Unknown9: byte;
    Str, StrPercent, Int, Wis, Dex, Con, Cha: byte;
    Morale, MoraleBreak: byte;
    RacialEnemy: byte;
    MoraleRecoveryTime: byte;
    Unknown10: array [1..3] of byte;
    SpecialistMage: word;
    ScriptOverride, ScriptClass, ScriptRace: TResRef;
    ScriptGeneral, ScriptDefault: TResRef;
  end;

const
  CRE_DUAL_MAGE    = 8;
  CRE_DUAL_FIGHTER = $10;
  CRE_DUAL_CLERIC  = $20;
  CRE_DUAL_THIEF   = $40;
  CRE_DUAL_DRUID   = $80;
  CRE_DUAL_RANGER  = $100;
  CRE_DUAL_ALL     = $1F8;

type
  TCREAIRec = record
    AI_EA, AI_General: byte;
    AI_Race, AI_Class, AI_Specific: byte;
    Gender: byte;
    Unknown13: array [1..5] of byte;
    Alignment: byte;
    Unknown14: array [1..4] of byte;
    DeathVar: array [0..31] of char;
  end;

  TCREOffsets1Rec = record
    KnownSpellOffset, KnownSpellCount: LongInt;
    MemorizeOffset, MemorizeCount: LongInt;
    MemorizedSpellOffset, MemorizedSpellCount: LongInt;
  end;

  TCREOffsets2Rec = record
    ItemSlotOffset: LongInt;
    ItemOffset, ItemCount: LongInt;
    EffectOffset, EffectCount: LongInt;
    Dialog: TResRef;
  end;

  // between TCRECommon1Rec and TCRECommon2Rec
  TCRE12ExtRec = record
    Unknown11: array [1..76] of byte;
    RaceName: array [0..31] of char;
    Unknown12: array [1..54] of byte;
    AI_Team, AI_Faction: byte;
  end;

  TCRE90ExtRec = record
    Unknown1: array [1..14] of byte;
    Var1, Var2: array [0..31] of char;
    Unknown2: array [1..26] of byte;
  end;

  TCREItemRec = record   // also used in ARE files
    ItemID: TResRef;
    Unknown1: word;
    Count: SmallInt;
    Count2: SmallInt;
    Count3: SmallInt;
    Flags: LongInt;
  end;

  TCREKnownSpellRec = record
    SpellID: TResRef;
    Level: word;
    SpellType: word;
  end;

  TCREMemorizedSpellRec = record
    SpellID: TResRef;
    Available: word;
    Unknown: word;
  end;

// -- SPL --------------------------------------------------------------------

  TSpellHeaderRec = record
    Signature: array [0..7] of char;
    TLKName: LongInt;
    Unknown1: LongInt;
    SoundID: TResRef;
    Unknown7: LongInt;
    SpellType: word;
    Unknown2: array [1..22] of byte;
    Level: integer;
    Unknown3: word;
    SpellIcon: TResRef;
    Unknown4: array [1..14] of byte;
    TLKDesc: LongInt;
    Unknown5: LongInt;
    SoundID2: TResRef;
    Unknown6: LongInt;
    AbilityOffset: LongInt;
    AbilityCount: word;
    EffectOffset: LongInt;
    EffectCount: LongInt;
  end;

// -- WMAP -------------------------------------------------------------------

  TWorldMapHeaderRec = record
    Signature: array [0..7] of char;
    WorldMapCount: integer;
    WorldMapOffset: LongInt;
  end;

  TWorldMapRec = record
    MapResName: TResRef;
    XSize, YSize: integer;
    Unknown: array [1..4] of integer;
    AreaCount, AreaOffset: integer;
    LinkOffset, LinkCount: integer;
    MapIconResName: TResRef;
    Unknown2: array [1..32] of integer;
  end;

  TWorldMapLinkPtrRec = record
    FirstLink: integer;
    LinkCount: integer;
  end;

  TWorldMapAreaRec = record
    AreaId1, AreaId2: TResRef;
    AreaName: array [0..31] of char;
    Flags: LongInt;
    BAMFrame: integer;
    X, Y: integer;
    TLKIndex1, TLKIndex2: integer;
    LoadScreen: TResRef;
    Links: array [1..4] of TWorldMapLinkPtrRec;
    Unknown2: array [1..32] of integer;
  end;

  TWorldMapLinkRec = record
    ToAreaIndex: integer;
    EntryPoint: array [0..31] of char;
    TravelHours: integer;
    Unknown: array [1..44] of integer;
  end;

const
  wmaVisible  = 1;
  wmaCanVisit = 4;
  wmaVisited  = 8;

// -- CHUI -------------------------------------------------------------------

type
  TChUIHeaderRec = record
    Signature: array [0..7] of char;
    PanelCount: integer;
    ControlIndexOffset: integer;
    PanelOffset: integer;
  end;

  TChUIControlIndexRec = record
    Offset: integer;
    Size: integer;
  end;

  TChUIPanelRec = record
    PanelID: integer;
    X, Y: word;
    Width, Height: word;
    HasBackground: word;
    ControlCount: word;
    BackgroundMOS: TResRef;
    FirstControl: word;
    Unknown: word;
  end;

  TChUIButtonRec = record
    BAMName: TResRef;
    BAMSeq: word;
    UpFrame, DownFrame: word;
    SelectedFrame, DisabledFrame: word;
  end;

  TChUISliderRec = record
    MOSBackground: TResRef;
    BAMKnob: TResRef;
    BAMKnobSeq: word;
    UngrabbedFrame, GrabbedFrame: word;
    Unknown: array [1..8] of word;
  end;

  TChUITextEditRec = record
    MOSFile1, MOSFile2, MOSFile3: TResRef;
    BAMCursor: TResRef;
    Unknown: array [1..12] of byte;
    BAMFont: TResRef;
    Unknown2: array [1..40] of byte;
  end;

  TChUITextAreaRec = record
    BAMFont1, BAMFont2: TResRef;
    Color1, Color2, Color3: integer;
    ScrollBarControlID: word;
  end;

  TChUITextRec = record
    TextStrref: integer;
    BAMFont: TResRef;
    Color1, Color2: integer;
    Unknown: word;
  end;

  TChUIScrollBarRec = record
    BAMFile: TResRef;
    BAMSeq: word;
    BAMUpUnpressedFrame, BAMUpPressedFrame: word;
    BAMDownUnpressedFrame, BAMDownPressedFrame: word;
    BAMTroughFrame, BAMSliderFrame: word;
    TextAreaControlID: word;
  end;

  TChUIControlType = (ctButton, ctUnknown1, ctSlider, ctEdit, ctUnknown4, ctTextArea,
      ctStatic, ctScrollBar);

  TChUIControlRec = record
    ControlID: integer;
    X, Y: word;
    Width, Height: word;
    ControlType: TChUIControlType;
    Unknown: byte;
    case TChUIControlType of
      ctButton: (Btn: TChUIButtonRec);
      ctSlider: (Slider: TChUISliderRec);
      ctEdit: (TextEdit: TChUITextEditRec);
      ctTextArea: (TextArea: TChUITextAreaRec);
      ctStatic: (Text: TChUITextRec);
      ctScrollBar: (ScrollBar: TChUIScrollBarRec);
  end;

// -- STOR -------------------------------------------------------------------

  TStoreHeaderRec = record
    Signature: array [0..7] of char;
    StoreType: LongInt;
    NameStrref: LongInt;
    Flags: LongInt;
    SellPrice, BuyPrice: LongInt;
    BuyPriceReduction: LongInt;
    StealDifficulty: LongInt;
    Unknown: array [1..8] of byte;
    PurchasedItemsOffset, PurchasedItemsCount: LongInt;
    SoldItemsOffset, SoldItemsCount: LongInt;
    IdentifyLore: LongInt;
    Unknown2: LongInt;
    RumorDlg: TResRef;
    DrinksOffset, DrinksCount: LongInt;
    DonateRumorDlg: TResRef;
    RoomTypes: LongInt;
    RoomPrices: array [1..4] of LongInt;  // Peasant, Merchant, Noble, Royal
    SpellsOffset, SpellsCount: LongInt;
    Unknown4: array [1..36] of byte;
  end;

  TStore10ItemRec = record
    Base: TCREItemRec;
    NumberInStock: integer;
    Flags: LongInt;
  end;

  TStore11ExItemRec = record
    Unknown: array [1..60] of byte;
  end;

  TStoreDrinkRec = record
    Unknown: array [0..7] of byte;
    NameStrref: LongInt;
    Price: LongInt;
    RumorChance: LongInt;
  end;

  TStoreSpellRec = record
    SpellID: TResRef;
    SpellPrice: LongInt;
  end;

// -- GAME -------------------------------------------------------------------

type
  TGameHeaderCommonRec = record
    Signature: array [0..7] of char;
    GameTime: LongInt;
    Unknown: array [1..6] of word;
    Gold: LongInt;
    Unknown2: LongInt;
    PartyOffset, PartyCount: LongInt;
    Unknown3: array [1..2] of LongInt;
    NPCOffset, NPCCount: LongInt;
    VarOffset, VarCount: LongInt;
    AreaResRef: TResRef;
    Unknown4: LongInt;
    JournalCount, JournalOffset: LongInt;
  end;

  TGameHeaderRec = record
    HdrCommon: TGameHeaderCommonRec;
    Unknown5: LongInt;
    AreaResRef2: TResRef;
    Unknown6: array [1..21] of LongInt;
  end;

  TGameHeaderTormentRec = record
    HdrCommon: TGameHeaderCommonRec;
    Unknown5: LongInt;
    Unknown6: LongInt;
    AreaResRef2: TResRef;
    KillVarOffset, KillVarCount: LongInt;
    Unknown7: LongInt;
    AreaResRef3: TResRef;
    Unknown8: array [1..16] of LongInt;
  end;

  TGameCharacterCommonRec = record
    Unknown: word;
    PartyPosition: SmallInt;
    CREOffset, CRESize: LongInt;
    Unknown2: array [1..2] of LongInt;
    Facing: LongInt;
    CurArea: TResRef;
    XCoord, YCoord: word;
    ViewRectX, ViewRectY: word;
    Unknown3: word;
    Happiness: word;
    Unknown4: array [1..37] of LongInt;
  end;

  TGameCharacterStatsRec = record
    MostPowerfulStrref: LongInt;
    MostPowerfulXP: LongInt;
    Unknown: array [1..3] of LongInt;
    KillXPChapter, KillCountChapter: LongInt;
    KillXPGame, KillCountGame: LongInt;
    FavSpells: array [1..4] of TResRef;
    FavSpellCount: array [1..4] of word;
    FavWeapons: array [1..4] of TResRef;
    FavWeaponCount: array [1..4] of word;
  end;

  TGameCharacterRec = record
    Common: TGameCharacterCommonRec;
    Name: array [0..31] of char;
    Unknown: LongInt;
    Stats: TGameCharacterStatsRec;
    Unknown2: array [0..7] of char;
    SoundSet: array [0..31] of char;
  end;

  TGameCharacterBaldurRec = record
    Common: TGameCharacterCommonRec;
    Name: array [0..31] of char;
    Unknown: LongInt;
    Stats: TGameCharacterStatsRec;
    SoundSet: array [0..7] of char;
  end;

  TGameCharacterTormentRec = record
    Common: TGameCharacterCommonRec;
    Unknown: array [1..11] of LongInt;
    Stats: TGameCharacterStatsRec;
    Unknown2: array [1..2] of LongInt;
  end;

  TGameCharacterIcewind2Rec = record
    Common: TGameCharacterCommonRec;
    Unknown: array [1..63] of LongInt;
    Unknown2: word;
    Name: array [0..31] of char;
    Unknown3: LongInt;
    Stats: TGameCharacterStatsRec;
    Unknown4: array [0..7] of char;
    SoundSet: array [0..31] of char;
    Unknown5: array [1..48] of integer;
    Unknown6: word;
  end;

  TGameVarRec = record
    Name: array [0..31] of char;
    Unknown1, Unknown2: LongInt;
    Value: LongInt;
    Unknown3: array [1..10] of LongInt;
  end;

  TGameJournalRec = record
    TextStrref: integer;
    Time: integer;
    Chapter: word;
    Unknown: word;
  end;

function Effect2Name (E: Integer; GameType: TGameType): string;   // from Petr Zahradnik's Expl

implementation

uses SysUtils;

function Effect2Name (E: Integer; GameType: TGameType): string;   // from Petr Zahradnik's Expl
begin
   case E of
     $00: Result:='AC';
     $01: Result:='ATTACKNO';
     $02: Result:='AWAKEN';
     $03: Result:='BERSERK';
     $04: Result:='CALM';
     $05: Result:='CHARM';
     $06: Result:='CHR';
     $07: Result:='COLORCHANGE';
     $08: Result:='COLORGLOW SOLID';
     $09: Result:='COLORGLOW PULSE';
     $0A: Result:='CON';
     $0B: Result:='CUREPOISON';
     $0C: Result:='DAMAGE';
     $0D: Result:='DEATH';
     $0E: Result:='DEFROST';
     $0F: Result:='DEX';
     $10: Result:='HASTE';
     $11: Result:='HEAL';
     $12: Result:='HITPOINTS';
     $13: Result:='INT';
     $14: Result:='INVISIBLE';
     $15: Result:='LORE';
     $16: Result:='LUCK';
     $17: Result:='MORALE';
     $18: Result:='PANIC';
     $19: Result:='POISON';
     $1A: Result:='REMOVECURSE';
     $1B: Result:='RESISTACID';
     $1C: Result:='RESISTCOLD';
     $1D: Result:='RESISTELECTRICITY';
     $1E: Result:='RESISTFIRE';
     $1F: Result:='RESISTMAGIC';
     $20: Result:='RESURRECT';
     $21: Result:='Save (Paralyze/Poison/Death)';
     $22: Result:='Save (Rod/Staff/Wand)';
     $23: Result:='Save (Petrify/Polymorph)';
     $24: Result:='Save (Breath)';
     $25: Result:='Save (Spell)';
     $26: Result:='SILENCE';
     $27: Result:='SLEEP';
     $28: Result:='SLOW';
     $29: Result:='SPARKLE';
     $2A: Result:='SPELLMEMORIZATIONMAGE';
     $2B: Result:='STONETOFLESH';
     $2C: Result:='STR';
     $2D: Result:='STUN';
     $2E: Result:='UNSTUN';
     $2F: Result:='VISIBLE';
     $30: Result:='VOCALIZE';
     $31: Result:='WIS';
     $32: Result:='SINGLECOLORPULSEALL';
     $33: Result:='COLORTINT_SOLID';
     $34: Result:='COLORLIGHT_SOLID';
     $35: Result:='ANIMATION_CHANGE';
     $36: Result:='THAC0';
     $37: Result:='SLAY';
     $38: Result:='ALIGNMENTREVERSAL';
     $39: Result:='ALIGNMENTCHANGE';
     $3A: Result:='DISPELEFFECTS';
     $3B: Result:='SKILLSTEALTH';
// V- misplaced
     $3C: Result:='CASTINGFAILURE';
//     $3D: Result:='SPELLMEMORIZATIONCLERIC';
// ^- misplaced somewhere
     $3E: Result:='SPELLMEMORIZETIONCLERIC';
//     $3E: Result:='Extra spells ?';
//     $3F: Result:='Infravision';

     $3F: Result:='INFRAVISION';
     $40: Result:='INFRAVISIONOFF';
     $41: Result:='BLUR';
     $42: Result:='TRANSCULENT';
     $43: Result:='SUMMON';
     $44: Result:='UNSUMMON';
     $45: Result:='NONDETECTION';
     $46: Result:='ENDNONDETECTION';
     $47: Result:='SEXCHANGE'; {47}
     $48: Result:='AICHANGE';
     $49: Result:='DAMAGEMOD';
     $4A: Result:='BLINDNESS';
     $4B: Result:='CUREBLINDESS';
     $4C: Result:='FEEBLEMINDEDNESS';
     $4D: Result:='CUREFEEBLMINDEDNESS';
     $4E: Result:='DISEASE';
     $4F: Result:='CUREDISEASE';
     $50: Result:='DEAFNESS';
     $51: Result:='CUREDEAFNESS';
     $52: Result:='SETAISCRIPT';
     $53: Result:='IMMUNITYTOPROJECTILE';
     $54: Result:='RESISTMAGICFIRE';
     {with tabs}
     $55: Result:='RESISTMAGICCOLD';
     $56: Result:='RESISTSLASHING';
     $57: Result:='RESISTCRUSHING';
     $58: Result:='RESISTPIERCING';
     $59: Result:='RESISTMISSILE';
     $5A: Result:='SKILLLOCKPICKING';
     $5B: Result:='SKILLTRAPS';
     $5C: Result:='SKILLPICKPOCKET';
     $5D: Result:='FATIGUE';
     $5E: Result:='INTOXICATION';
     $5F: Result:='SKILLTRACKING';
     $60: Result:='LEVEL';
     $61: Result:='STREXTRA';
     $62: Result:='REGENERATION';
     $63: Result:='DURATIONMOD';
     $64: Result:='PROTECTIONFROMCREATURE';
{65} $65: Result:='IMMUNITYTOEFFECT';   {petrification,poison,feeblemindnes,confusion,fear,charm...}
     $66: Result:='IMMUNITYTOSPELLLEVEL';
     $67: Result:='NAME';
     $68: Result:='XP';
     $69: Result:='GOLD';
     $6A: Result:='MORALEBREAK';
     $6B: Result:='PORTRAIT';
     $6C: Result:='REPUTATION';
     $6D: Result:='HOLD_CREATURE';
     $6E: Result:='RETREAT_FROM';
     $6F: Result:='CREATE_WEAPON';
     $70: Result:='DESTROY_WEAPON';
     $71: Result:='EQUIP_WEAPON';
     $72: Result:='DITHER';
     $73: Result:='DETECT_ALIGNMENT';
     $74: Result:='DISPEL_INVISIBLE';
     $75: Result:='SHOW_AREA';
     $76: Result:='SHOW_CREATURES';
     $77: Result:='MIRROR_IMAGE';
     $78: Result:='IMUNITY_TO_WEAPON';
     $79: Result:='VISUAL_ANIMATION_EFFECT';
{7A} $7A: Result:='CREATE_ITEM';
     $7B: Result:='DESTROY_ITEM';
     $7C: Result:='TELEPORT';
     $7D: Result:='KNOCK';
     $7E: Result:='MOVEMENT_RATE';
     $7F: Result:='RANDOM_SUMMON';
     $80: Result:='CONFUSION';
     $81: Result:='NON_CUMULATIVE_AID';
     $82: Result:='NON_CUMULATIVE_BLESS';
     $83: Result:='NON_CUMULATIVE_CHANT';
     $84: Result:='NON_CUMULATIVE_DRAW_UPON_HOLY_MIGHT';
     $85: Result:='NON_CUMULATIVE_LUCK';
{86} $86: Result:='PETRIFICATION';
     $87: Result:='POLYMORPH';
     $88: Result:='FORCE_VISIBLE';
     $89: Result:='NON_CUMULATIVE_CHANTBAD';
     $8A: Result:='SETSEQUENCE';
     $8B: Result:='DISPLAYSTRING';
     $8C: Result:='CASTINGGLOW';
     $8D: Result:='VISUALSPELLHIT';
{}   $8E: Result:='PORTRAITICON';
     $8F: Result:='REPLACE_ITEM';
     $90: Result:='DISABLE_BUTTON';
     $91: Result:='DISABLE_SPELLTYPE';
     $92: Result:='CAST_SPELL';
     $93: Result:='LEARN_SPELL';
     $94: Result:='CAST_SPELL_POINT';
     $95: Result:='IDENTIFY';
     $96: Result:='DETECTTRAPS';
     $97: Result:='REPLACESELF';
     $98: Result:='PLAYMOVIE';
     $99: Result:='SANCTUARY';
     $9A: Result := 'Entangle';
     $9B: Result := 'Minor Globe';
     $9C: Result := 'Prot.Normal Missiles';
     $9D: Result := 'Web';
     $9E: Result := 'Grease';
     $9F: Result := 'Mirror Image (2)';
     $A4: Result := 'Slow Poison';
     $A5: Result := 'Burning Hands';
     $A6: Result := 'Magic Resistance (2)';
     $A8: Result := 'Remove Creature';
     $A9: Result := 'Remove Icon';
     $AB: Result := 'Give Innate Ability';
     $AC: Result := 'Remove Spell';
     $AE: Result:='Make Sound';
     $AF: Result := 'Hold Creature Type';
     $B1: Result := 'Damage vs Creature Type';
     $B9: Result := 'Hold Creature (2)';
     $BC: Result := 'Increase Spellcasting Speed';
     $BD: Result := 'Spell Speed Factor';
     $BE: Result := 'Attack Speed Factor';
     else begin
       if GameType = gtIcewind then begin
         case E of
           $E9: Result := 'Effect Animation';
           $105: Result := 'Immunity to Effect';
           else
             Result:=Format('Effect: 0x%x',[E]);
         end;
       end
       else
         Result:=Format('Effect: 0x%x',[E]);
     end;
   end;
end;

end.
